structure Core =
struct

open Syntax

(* ------------------------   EVALUATION  ------------------------ *)

exception NoRuleApplies

fun isnumericval t =
    case t
      of TmZero _ => true
       | TmSucc (_,t1) => isnumericval t1
       | _ => false

fun isval t =
    case t
      of TmTrue _ => true
       | TmFalse _ => true
       | t => isnumericval t

(* eval1:  "call by value" single-step evaluation *)
fun eval1 t =
    case t
      of TmIf(_,TmTrue _,t2,t3) => t2
       | TmIf(_,TmFalse _,t2,t3) => t3
       | TmIf(info,t1,t2,t3) => TmIf(info,eval1 t1, t2, t3)
       | TmSucc(info,t1) => TmSucc(info,eval1 t1)
       | TmPred(_,TmZero info) => TmZero NONE  (* predecessor of zero is zero *)
       | TmPred(info,TmSucc(_,t1)) => 
	   if (isnumericval t1) then t1
	   else TmPred(info,TmSucc(NONE,eval1 t1))
       | TmPred(info,t1) => TmPred(info,eval1 t1)
       | TmIsZero(_,TmZero _) => TmTrue NONE
       | TmIsZero(info1,TmSucc(info2,t1)) =>
	   if (isnumericval t1) then TmFalse NONE
           else TmIsZero(info1,TmSucc(info2,eval1 t1))
       | TmIsZero(info,t1) => TmIsZero(info,eval1 t1)
       | _ => raise NoRuleApplies

fun eval t =
     let val t' = eval1 t
      in eval t'
     end 
     handle NoRuleApplies => t

end (* structure Core *)
